#!/usr/bin/env python
"""Package cogent3.format: provides modules for writing specific file formats.

Currently provides:
    xml:  xml base class
    file: general functions to read and write files
"""
__all__ = [
    "alignment",
    "bedgraph",
    "clustal",
    "fasta",
    "gde",
    "nexus",
    "paml",
    "phylip",
    "table",
]
