#!/usr/bin/env python

__all__ = [
    "test_alignment",
    "test_alphabet",
    "test_annotation",
    "test_core_standalone",
    "test_features",
    "test_genetic_code",
    "test_info",
    "test_location",
    "test_maps",
    "test_moltype",
    "test_profile",
    "test_sequence",
    "test_tree",
]
